using System;

namespace Equals
{
	class Program
	{
		static void Main(string[] args)
		{
			Model first1 = new Model(1);
			Model second1 = new Model(1);

			Console.WriteLine("{0} == {1} is {2}", first1, second1, IsTrue(first1 == second1));
			Console.WriteLine("{0} == {1} is {2}", second1, first1, IsTrue(second1 == first1));
			
			TestNotGenerice(first1, second1);
			TestGeneric(first1, second1);
			TestModel(first1, second1);
			TestGenericWithInterface(first1, second1);

			Console.Read();
		}

		private static void TestNotGenerice(Model first1, Model second1)
		{
			Console.WriteLine("{0} == {1} is {2}", first1, second1, IsTrue(first1 == second1));
		}
		
		private static void TestGeneric<T>(T first1, T second1) where T:class 
		{
			Console.WriteLine("{0} == {1} is {2}", first1, second1, IsTrue(first1 == second1));
		}

		private static void TestModel<T>(T first1, T second1) where T : Model
		{
			Console.WriteLine("{0} == {1} is {2}", first1, second1, IsTrue(first1 == second1));
		}

		private static void TestGenericWithInterface<T>(T first1, T second1) where T : class, IEquatable<T>
		{
			Console.WriteLine("{0} == {1} is {2}", first1, second1, IsTrue(first1 == second1));
		}

		private static string IsTrue(bool value)
		{
			return (value) ? "PASS" : "FAIL";
		}
	}

	class Model : IEquatable<Model>
	{
		private readonly int _id;

		public Model(int _id)
		{
			this._id = _id;
		}

		protected int id
		{
			get { return _id; }
		}

		public override bool Equals(object obj)
		{
			if (ReferenceEquals(null, obj))
				return false;

			if (ReferenceEquals(this, obj))
				return true;

			Model m = obj as Model;
			if (m == null)
				return false;

			return Equals(m);
		}
		
		public static bool operator ==(Model x, Model y)
		{
			if (ReferenceEquals(x, y))
				return true;
			if (ReferenceEquals(x, null))
				return false;

			return x.Equals(y);
		}

		public static bool operator !=(Model x, Model y)
		{
			return !(x == y);
		}

		public override int GetHashCode()
		{
			return _id.GetHashCode();
		}

		public override string ToString()
		{
			return _id.ToString();
		}

		public virtual bool Equals(Model other)
		{
			if (ReferenceEquals(null, other))
				return false;

			if (ReferenceEquals(this, other))
				return true;

			return _id.Equals(other._id);;
		}
	}
}
